#!/usr/bin/env python3

# Copyright (c) 2019 Christian Barral

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import datetime
from typing import Union

from . import utils
from .elements import Host


OUTPUT_FORMATS = ['normal', 'grep', 'xml']

class NmapScanResult:
    """ An instance of this class encapsulates the output of a Nmap
    execution.

    Scan results have several parts that should be distinguished. First,
    it contains overall scan information, which includes the time spent,
    nmap version, arguments that have been executed, etc...

    Then, the results have information from the scanned hosts, including
    general information from the host: IP Address, hostnames, scanned ports, services
    inside those ports... As well as scripts. These scripts can be host-oriented
    which are, indeed, binded to the host, and service-oriented, which are binded to
    the services.

    Additionally, these hosts could also have Operating System results and NSE
    results, which are unique from this library.
    """

    __slots__ = ('_scanner', '_arguments', '_start_timestamp', '_start_datetime',
                 '_version', '_end_timestamp', '_end_datetime', '_elapsed', '_summary',
                 '_exit_status', '_hosts_up', '_hosts_down', '_num_hosts', '_scan_info',
                 '_verbose', '_debug', '_hosts', '_tolerant_errors', '_xml_output', 
                 '_grep_output', '_normal_output', '_index')

    def __init__(self, **kwargs):
        self.scanner = kwargs.get('scanner', None)
        self.arguments = kwargs.get('arguments', None)
        self.start_timestamp = kwargs.get('start_timestamp', None)
        self.start_datetime = kwargs.get('start_datetime', None)
        self.version = kwargs.get('version', None)
        self.end_timestamp = kwargs.get('end_timestamp', None)
        self.end_datetime = kwargs.get('end_datetime', None)
        self.elapsed = kwargs.get('elapsed', None)
        self.summary = kwargs.get('summary', None)
        self.exit_status = kwargs.get('exit_status', None)

        self._hosts_up = kwargs.get('hosts_up', None)
        self._hosts_down = kwargs.get('hosts_down', None)
        self._num_hosts = kwargs.get('num_hosts', None)
    
        self.scan_info = kwargs.get('scan_info', None)
        self.verbose = kwargs.get('verbose', None)
        self.debug = kwargs.get('debug', None)

        self.tolerant_errors = None
        
        self._hosts = []

        self._xml_output = None
        self._grep_output = None
        self._normal_output = None

        self._index = -1

    @property
    def scanner(self) -> Union[None,str]:
        """ Nmap scanner information
        """
        return self._scanner
    
    @scanner.setter
    def scanner(self, v):
        assert v is None or isinstance(v, str), 'NmapScanResult.scanner must be None or str'

        self._scanner = v
    
    @property
    def arguments(self) -> Union[None,str]:
        """ Command-line arguments used
        """
        return self._arguments
    
    @arguments.setter
    def arguments(self, v):
        assert v is None or isinstance(v, str), 'NmapScanResult.arguments must be None or str'

        self._arguments = v

    @property
    def start_timestamp(self) -> Union[None,int]:
        """ Start time timestamp
        """
        return self._start_timestamp
    
    @start_timestamp.setter
    def start_timestamp(self, v):
        assert v is None or isinstance(v, str) or isinstance(v, int), 'NmapScanResult.start_timestamp must be None, str or int'

        if v is not None:
            self._start_timestamp = int(v)
        else:
            self._start_timestamp = v   
    
    @property
    def start_datetime(self) -> Union[None,datetime.datetime]:
        """ Start time datetime object
        """
        return self._start_datetime
    
    @start_datetime.setter
    def start_datetime(self, v):
        
        if v is not None:
            self._start_datetime = datetime.datetime.fromtimestamp(int(v))
        else:
            self._start_datetime = None

    @property
    def version(self) -> Union[None,str]:
        """ Get Nmap version
        """
        return self._version
    
    @version.setter
    def version(self, v):
        assert v is None or isinstance(v, str), 'NmapScanResult.start_timestamp must be None or str'

        self._version = v

    @property
    def end_timestamp(self) -> Union[None,int]:
        """ End time timestamp
        """
        return self._end_timestamp
    
    @end_timestamp.setter
    def end_timestamp(self, v):
        assert v is None or isinstance(v, (str, int)), 'NmapScanResult.end_timestamp must be None, str or int'

        if v is not None:
            self._end_timestamp = int(v)
        else:
            self._end_timestamp = v
    
    @property
    def end_datetime(self) -> Union[None,datetime.datetime]:
        """ End time datetime object
        """
        return self._end_datetime
    
    @end_datetime.setter
    def end_datetime(self, v):
        if v is not None:
            self._end_datetime = datetime.datetime.fromtimestamp(int(v))
        else:
            self._end_datetime = None

    @property
    def elapsed(self) -> Union[None,float]:
        """ Elapsed time
        """
        return self._elapsed
    
    @elapsed.setter
    def elapsed(self, v):
        assert v is None or isinstance(v, str) or isinstance(v, float), 'NmapScanResult.elapsed must be None, str or float'
        try:
            self._elapsed = float(v)
        except ValueError:
            pass
    
    @property
    def summary(self) -> Union[None,str]:
        """Scan summary
        """
        return self._summary
    
    @summary.setter
    def summary(self, v):
        assert v is None or isinstance(v, str), 'NmapScanResult.summary must be None or str'
        
        self._summary = v
    
    @property
    def exit_status(self) -> Union[None,str]:
        """ Nmap exit status
        """
        return self._exit_status
    
    @exit_status.setter
    def exit_status(self, v):
        assert v is None or isinstance(v, str), 'NmapScanResult.exit_status must be None or str'
        
        self._exit_status = v

    @property
    def hosts_up(self) -> Union[None,int]:
        """ Number of hosts up
        """
        return self._hosts_up
    
    @property
    def hosts_down(self) -> Union[None,int]:
        """ Number of hosts down
        """
        return self._hosts_down
    
    @property
    def num_hosts(self) -> Union[None,int]:
        """ Total number of hosts
        """
        return self._num_hosts

    @property
    def scan_info(self) -> Union[None,str]:
        """ Scan information
        """
        return self._scan_info

    @scan_info.setter
    def scan_info(self, v):
        assert v is None or isinstance(v, dict), 'NmapScanResult.scan_info must be None or a dictionary'

        if v is None:
            self._scan_info = {}
        else:
            self._scan_info = v

    @property
    def verbose(self) -> Union[None,int]:
        """ Verbosity level
        """
        return self._verbose

    @verbose.setter
    def verbose(self, v):
        assert v is None or isinstance(v, (int, str)), 'NmapScanResult.verbose must be None, int or str'
        
        if v is not None:
            self._verbose = int(v)
        else:
            self._verbose = None

    @property
    def debug(self) -> Union[None,int]:
        """ Debugging level
        """
        return self._debug

    @debug.setter
    def debug(self, v):
        assert v is None or isinstance(v, (int, str)), 'NmapScanResult.debug must be None, int or str'
        
        if v is not None:
            self._debug = int(v)
        else:
            self._debug = None

    @property
    def tolerant_errors(self) -> Union[None,str]:
        """ String containing tolerant errors
        """
        return self._tolerant_errors
    
    @tolerant_errors.setter
    def tolerant_errors(self, v):
        assert v is None or isinstance(v, str), 'NmapScanResult.tolerant_errors must be None or str'

        self._tolerant_errors = v

    def __len__(self):
        return len(self._hosts)
    
    def __getitem__(self, v):
        """ Flexible get item by position, ip, ip ranges, hostnames and any combination of the last three.

        If an integer is specified, then it will return a host in position v, acting as a normal list.
        If an str is specified, then it will check if the provided string has spaces (which indicates that the user is requesting multiple hosts)
        and if so, it will do a single search for all of them.
        If tuple is set, then the user has specified multiple arguments separated by commas, so it will act the same as handling strings with spaces.

        If the user specifies a multi-host value, then the resposne will be a list, empty or not.
        If the user specifies an integer, or a single host to be return, then the method will return None or the host, if it exists.

        :param v: Value, or values, to retrive from the hosts list.
        """

        if isinstance(v, int):
            return self._hosts[v]
        elif isinstance(v, (str, tuple)):
            to_return = []
            multi_return = False
            if isinstance(v, str):
                v = v.strip()
                if ' ' in v:
                    v = [x for x in v.split() if v]
                else:
                    v = [v]

            if len(v) > 1:
                multi_return = True

            for i in v:
                ips = []
                hostnames = []
                if utils._SINGLE_IP_ADDRESS_REGEX.fullmatch(i):
                    ips.append(i)
                elif utils._IP_ADDRESS_WITH_CIDR_REGEX.fullmatch(i):
                    ips.extend(utils.dispatch_network(i))
                elif utils._IP_RANGE_REGEX.fullmatch(i):
                    split_value = i.split('-')
                    ips.extend(utils.ip_range(split_value[0], split_value[1]))
                elif utils._PARTIAL_IP_RANGE_REGEX.fullmatch(i):
                    ips.extend(utils.partial_ip_range(i))
                else:
                    hostnames.append(i)
            
            
            for host in self._hosts:
                if host.ipv4 in ips or [i for i in host.hostnames() if i in hostnames]:
                    to_return.append(host)
            
            if not multi_return:
                try:
                    return to_return[0]
                except IndexError:
                    return None
            
            else:
                return to_return

        else:
            raise TypeError('Invalid index type. Must be int, str or tuple but found {}'.format(type(v)))

    def __next__(self):
        """ Defines the iterator behaviour
        """
        if (self._index + 1) < len(self._hosts):
            self._index += 1
            return self._hosts[self._index]
        else:
            raise StopIteration

    def __iter__(self):
        """ Return an iterator for the hosts
        """

        return iter(self._hosts)

    def __len__(self):
        """ Get the length of the result object
        """

        return len(self._hosts)

    def __contains__(self, v):
        """ Check if an element v is in result
        """
        is_ip = False
        if utils._SINGLE_IP_ADDRESS_REGEX.fullmatch(v):
            is_ip = True

        for i in self:
            if (is_ip and i.ipv4 == v) or (not is_ip and v in i.hostnames()):
                return True

        return False

    def _add_hosts(self, *args):
        """ Add hosts objects to the current instance.

        :param args: Any number of Hosts instances to add
        :raises TypeError: If any of the instances is not from the Host class
        """

        for i in args:
            if not isinstance(i, Host):
                raise TypeError('Cannot add non-Host objects to a NmapScanResult')
            self._hosts.append(i)

    def scanned_hosts(self):
        """ Returns the hosts objects from the hosts that responded to the scan
        
        :return: List of hosts objects
        """

        return self._hosts
    
    def get_output(self, output_type):
        """ Returns, if any, the specified output format from 'xml', 'grep' or 'normal'.

        :returns: String representing the Nmap output from the specified output_type, or None if there is no one
        :raises ValueError: If the output_type has an invalid value.
        """

        if output_type not in OUTPUT_FORMATS:
            raise ValueError('Invalid output_type value. Expected on of the follwong: {}'.format(', '.join(OUTPUT_FORMATS)))
        
        if output_type == 'xml':
            return self._xml_output
        elif output_type == 'normal':
            return self._normal_output
        elif output_type == 'grep':
            return self._grep_output


#!/usr/bin/env python

# Copyright (c) 2019 Christian Barral

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re

from typing import Iterable, Union, Callable, get_type_hints

from .utils import targets_to_list, ports_to_list, extend_port_list
from .exceptions import EngineError, StopExecution
from .elements import Host, Port, Service


class _DelayedGlobalParserAbstraction:
    """ Represents a global parser that would be registered through OOP
    """
    def __init__(self, func) -> None:
        self.func = func


class _DelayedParserAbstraction(_DelayedGlobalParserAbstraction):
    """ Represents a parser that would be registered through OOP
    """
    def __init__(self, script_name, func) -> None:
        super().__init__(func)
        self.script_name = script_name


class NSEMeta(type):
    """ Metaclass to represent the behaivor for registering class methods as NSE scripts for Object-Oriented engines.

    It defines a set to store the functions for dalayed registry, which means, methods that will have to be added after instatiating the object
    subclassing NSE. It marks them with a flag (a protected attribute) for further registering them depending on the instance type.
    """
    def __init__(cls, name, bases, attrs):

        _delayed_registry_set = set()
        for _, method in attrs.items():
            if isinstance(method, property):
                method = method.fget
            if hasattr(method, '_delayed_registry'):
                _delayed_registry_set.add(getattr(method, '_delayed_registry'))

        @property
        def _delayed_registry(self):
            registries = _delayed_registry_set.copy()
            try:
                registries.update(super(cls, self)._delayed_registry_set)
            except AttributeError:
                pass
            return registries
        
        # This is used to propagate through the hierarchy
        cls._delayed_registry = _delayed_registry


def host_script(name: str, targets: Union[str,Iterable] = '*'):
    """ Decorator to be used from a class inheriting NSE, that automatically adds the decorated method as 
        an NSE host script with the specified name.
        
        :param name: Name of the function/script to be used later on to retrieve the information gathered by it.
        :param targets: Targets to be affected by the function. Specify them the same way as you specify scan targets. '*' would be all of them.
        :returns: Funciton decorator
    """
    def inner(f):
        f._delayed_registry = _NSEHostScript(name, f, targets, delayed=True)
        return f
    return inner


def port_script(name: str, port: Union[int,str,Iterable], targets: Union[str,Iterable] = '*', 
                proto: str = '*', states: Union[None,Iterable] = None):
    """ Decorator to be used from a class inheriting NSE, that registers the given function into the NSE as a port script.

        :param name: Name of the function/script to be used later on to retrieve the information gathered by it.
        :param port: Port(s) to be affected by the function. You can specify, '*' to target all ports. You can also set them the
                    same way as ports in the nmapthon2.scanner.NmapScanner.scan() method, but without the tcp(), udp() or top_ports() functions.
        :param targets: Targets to be affected by the function. Specify them the same way as you specify scan targets. '*' would be all of them.
        :param proto: Protocol of the port to be affected by the function. Default is '*', which applies to any protocol, but it can be either 'tcp' or 'udp'.
        :param states: List of states valid for function execution, can be a list with the following values in it: 'open', 'filtered' and/or 'closed'. By default, port scripts only target open ports
    """
    def inner(f):
        f._delayed_registry = _NSEPortScript(name, f, targets, port, proto, states)
        return f
    return inner


def parser(script_name: str):
    """ Decorator to be used from a class inheriting NSE, that registers the given function as a NSE parser for a given script
        
        :param name: Script name
        :returns: Function decorator
    """
    def inner(f):
        f._delayed_registry = _DelayedParserAbstraction(script_name, f)
        return f
    return inner

def global_parser(f):
    """Decorator to be used from a class inheriting NSE, that registers the given function as a NSE global parser
        
        :returns: Function docorator
    """
    def inner():
        f._delayed_registry = _DelayedGlobalParserAbstraction(f)
        return f
    return inner


class _NSEHostScript:
    """ An individual Python function that is executed as if it were an Nmap NSE host script.
    The script is represented by a name and a function to execute per alive host.

        :param name: Name of the function (PyNSEScript name)
        :param func: Function to execute
        :param targets: Filter to specify which targets should be affected by the function.
        :param delayed: Delayed allows the NSE object know that the script has been registered through a class, meaning 
                        that it will need to register after instantiation (delayed registry).
    """

    __slots__ = ('_name', '_func', '_targets', '_delayed')

    def __init__(self, name: str, func: Callable, targets: Union[str, Iterable], delayed: bool = False):
        self.name = name
        self.func = func
        self.targets = targets
        self.delayed = delayed

    @property
    def name(self):
        return self._name

    @property
    def func(self):
        return self._func

    @property
    def targets(self):
        return self._targets

    @property
    def delayed(self):
        return self._delayed

    @name.setter
    def name(self, v):
        self._name = v

    @func.setter
    def func(self, v):
        if not callable(v):
            raise EngineError('Function parameter is not callable: {}'.format(v))

        self._func = v

    @targets.setter
    def targets(self, v):
        if isinstance(v, list):
            new_list = []
            for i in v:
                new_list.extend(targets_to_list(i))
            self._targets = list(new_list)
        elif isinstance(v, str):
            if v.strip() == '*':
                self._targets = v
            else:
                self._targets = targets_to_list(v)
        else:
            raise EngineError('Invalid targets data type: {}'.format(type(v)))
    
    @delayed.setter
    def delayed(self, v):
        assert isinstance(v, bool), 'NSEHostScript.delayed must be of bool type'
        self._delayed = v

class _NSEPortScript(_NSEHostScript):
    """ Represents an NSE port script.
    As well as those scripts, it can have any number of arguments and it can assign output to the NmapScanResult object.
    The functions are only executed if the port or host (depending on the function type) is open,
    but the user can also specify which of the three port states ('open', 'filtered'or 'closed') are valid for the
    function to execute.
        :param ports: Port or ports to target.
        :param proto: Transport layer protocol ('tcp', 'udp' or '*' for both)
        :param states: List of states valid for function execution
    """

    __slots__ = ('_ports', '_proto', '_states')

    def __init__(self, name: str, func: Callable, targets: Union[str,Iterable], ports: Union[int,str,Iterable], proto: str, states: Union[str,Iterable]):
        super().__init__(name, func, targets)
        self.ports = ports
        self.proto = proto
        self.states = states

    @property
    def ports(self):
        return self._ports

    @property
    def proto(self):
        return self._proto

    @property
    def states(self):
        return self._states

    @ports.setter
    def ports(self, v):
        
        if v is None:
            self._ports = v
        elif isinstance(v, (int, str)):
            self._ports = ports_to_list(str(v))

        elif isinstance(v, Iterable):
            self._ports = extend_port_list(v)

        else:
            raise EngineError('Invalid ports data type: {}'.format(type(v)))

    @proto.setter
    def proto(self, v):
        if v is None:
            self._proto = None
        elif isinstance(v, str) and v.lower() in ['tcp', 'udp', '*']:
            self._proto = v.lower()
        else:
            raise EngineError('Invalid proto value: {} ({})'.format(v, type(v)))

    @states.setter
    def states(self, v):

        if v is None:
            self._states = ['open']

        elif not all(x in ['open', 'closed', 'filtered'] for x in v):
            raise EngineError('PyNSEScript states must be "open", "closed" or "filtered".')

        else:
            self._states = v


class NSE(metaclass=NSEMeta):
    """ Represents the NSE (Nmap Script Engine). It is used to instantiate an object that is passed to the
    NmapScanner __init__ or scan() methods and it registers new "Python NSE scripts", that are function written in Python. These
    functions execute depending on the conditions defined by the user, and they can be host or port oriented.
    Several decorators are offered to make it easy for the user to include new functions.

    NSE objects can also register parsers, which are Python functions that automatically parser Nmap scripts outputs. Global parsers are 
    always executed on every script output, while normal parses only parsers specifici scripts.
    """

    __slots__ = ('_global_parsers', '_parsers', '_host_scripts', '_port_scripts')


    def __init__(self):
        
        self._global_parsers = []
        self._parsers = {}

        self._host_scripts = []
        self._port_scripts = []

        for i in self._delayed_registry:
            if isinstance(i, _NSEPortScript):
                self._port_scripts.append(i)
            elif isinstance(i, _NSEHostScript):
                self._host_scripts.append(i)
            elif isinstance(i, _DelayedParserAbstraction):
                self._parsers[i.script_name] = getattr(self, i.func.__name__)
            elif isinstance(i, _DelayedGlobalParserAbstraction):
                self._global_parsers.append(getattr(self, i.func.__name__))
            else:
                raise EngineError('Could not add NSE script to engine. Unkown type: {}'.format(type(i)))

    def add_port_script(self, func: Callable, name: str, port: Union[int,str,Iterable], targets: Union[str,Iterable] = '*', 
                        proto: Union[None,str,Iterable] = '*', states: Union[None,Iterable] = None):
        """ Register a given function to execute on a given port.

        :param func: Function to register
        :param name: Name of the function/script to be used later on to retrieve the information gathered by it.
        :param port: Port(s) to be affected by the function. You can specify, '*' to target all ports. You can also set them the
                    same way as ports in the nmapthon2.scanner.NmapScanner.scan() method, but without the tcp(), udp() or top_ports() functions.
        :param targets: Targets to be affected by the function. Specify them the same way as you specify scan targets. '*' would be all of them.
        :param proto: Protocol of the port to be affected by the function. Default is '*', which applies to any protocol, but it can be either 'tcp' or 'udp'.
        :param states: List of states valid for function execution, can be a list with the following values in it: 'open', 'filtered' and/or 'closed'. By default, port scripts only target open ports
    
        """
        self._port_scripts.append(_NSEPortScript(name, func, targets, port, proto, states))

    def add_host_script(self, func: Callable, name: str, targets: Union[str,Iterable] = '*'):
        """ Register a given function to execute on a hosts

        :param func: Callback function to register
        :param name: Name of the function/script to be used later on to retrieve the information gathered by it.
        :param targets: Targets to be affected by the function. Asterik means all of them, but they can be specified the same way as you specify targets in the scan() method, including network ranges, partial ranges, etc...
        """
        self._host_scripts.append(_NSEHostScript(name, func, targets))

    def add_global_parser(self, callback: Callable):
        """ Adds a function to the global parsers.
        
        :param callback: Function to be executed to parse the output
        """
        self._global_parsers.append(callback)

    def add_parser(self, callback: Callable, script_name: str):
        """ Adds a function to the parsers for a given script name
        
        :param script_name: Name of the script to parse
        :param callback: Function to execute. Must accept one parameter, which will be the script output.
        :raises EngineError: Whenever the engine already has a registered function for the given script name.
        """
        if script_name in self._parsers:
            raise EngineError('"{}" already has a parsing function'.format(script_name))
        self._parsers[script_name] = callback

    def port_script(self, name: str, port: Union[int,str,Iterable], targets: Union[str,Iterable] = '*', 
                    proto: str = '*', states: Union[None,Iterable] = None):
        """ A decorator to register the given function into the NSE as a port script.

        :param name: Name of the function/script to be used later on to retrieve the information gathered by it.
        :param port: Port(s) to be affected by the function. You can specify, '*' to target all ports. You can also set them the
                    same way as ports in the nmapthon2.scanner.NmapScanner.scan() method, but without the tcp(), udp() or top_ports() functions.
        :param targets: Targets to be affected by the function. Specify them the same way as you specify scan targets. '*' would be all of them.
        :param proto: Protocol of the port to be affected by the function. Default is '*', which applies to any protocol, but it can be either 'tcp' or 'udp'.
        :param states: List of states valid for function execution, can be a list with the following values in it: 'open', 'filtered' and/or 'closed'. By default, port scripts only target open ports
        """

        def decorator(f):
            self.add_port_script(f, name, port, targets, proto, states or ['open'])
            return f

        return decorator

    def host_script(self, name: str, targets: Union[str,Iterable] = '*'):
        """ A decorator to register the given function into the NSE as a host script

        :param name: Name of the function/script to be used later on to retrieve the information gathered by it.
        :param targets: Targets to be affected by the function. Specify them the same way as you specify scan targets. '*' would be all of them.
        :returns: Function decorator
        """

        def decorator(f):
            self.add_host_script(f, name, targets)
            return f

        return decorator

    @staticmethod
    def global_parser(f):
        """ A decorator to register the given function as a NSE global parser
        
        :returns: Function docorator
        """

        def decorator():
            self.add_global_parser(f)
            return f
        
        return decorator
    
    def parser(self, name: str):
        """ A decorator to register the given function as a NSE parser for a given script
        
        :param name: Script name
        :returns: Function decorator
        """

        def decorator(f):
            self.add_parser(name, f)
            return f
        
        return decorator

    def _apply_host_scripts(self, host: Host) -> None:
        """ Execute all host scripts for a given host.

        :param host: Reference to a Host object
        """
        for i in self._host_scripts:
            if i.targets == '*' or host.ipv4 in i.targets or any(x for x in host.hostnames() if x in i.targets):
                try:
                    if i.delayed:
                        host._add_script(i.name, getattr(self, i.func.__name__)(host))
                    else:
                        host._add_script(i.name, i.func(host))
                except StopExecution:
                    pass
    

    def _apply_port_scripts(self, host: Host, port: Port, service: Service) -> None:
        """ Execute all port scripts for a given host, port and service.

        :param host: Reference to a Host object
        :param port: Reference to a port object
        :param service: Reference to a service object
        """

        if not self._port_scripts:
            return
        
        for i in self._port_scripts:
            if i.targets == '*' or host.ipv4 in i.targets or any(x for x in host.hostnames() if x in i.targets):
                if (i.proto == '*' or port.protocol == i.proto) and (i.ports == '*' or port.number in i.ports) and port.state in i.states:
                    try:
                        if i.delayed:
                            service._add_script(i.name, getattr(self, i.func.__name__)(host, port, service))
                        else:
                            service._add_script(i.name, i.func(host, port, service))
                    except StopExecution:
                        pass



#!/usr/bin/env python3

# Copyright (c) 2019 Christian Barral

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Offers an object-oriented representation of the scan status

import re

from .exceptions import XMLParsingError

import datetime
import xml.etree.ElementTree as ET


TASK_PROGRESS_REGEX = re.compile(r'<taskprogress .+\/>')
XML_TASK_EXTRACTOR_REGEX = re.compile(r'[a-z]+="[^"]+"')


class Status:
    """ Represents the Nmap scan status.
    
    The status has information from the task being performed, the timestamp, the completion percentage, the remaining
    number of hosts and the ETA.
    """

    __slots__ = ('_task', '_time', '_percent', '_remaining', '_etc')

    def __init__(self, **kwargs):
        self.task = kwargs.get('task', None)
        self.time = kwargs.get('time', None)
        self.percent = kwargs.get('percent', None)
        self.remaining = kwargs.get('remaining', None)
        self.etc = kwargs.get('etc', None)

    @classmethod
    def from_raw_xml(cls, xml):
        """ Creates a new instance from raw XML output

        :param xml: Raw utf8 decoded XML output
        :returns: New Status instance
        :raises XMLParsingError: If the output cannot be parsed to XML
        """
        try:
            last_status = TASK_PROGRESS_REGEX.findall(xml)[-1]
        except IndexError:
            return None
        else:
            status_info = {}
            for key_value_pair in XML_TASK_EXTRACTOR_REGEX.findall(last_status):
                split_values = key_value_pair.split('=')
                key = split_values[0]
                value = ''.join(split_values[1:]).strip('"')
                status_info[key] = value

            return cls(**status_info)

    @property
    def task(self):
        return self._task
    
    @task.setter
    def task(self, v):
        assert v is None or isinstance(v, str), 'Status.task can only be None or str'
        
        self._task = v

    @property
    def time(self):
        return self._time
    
    @time.setter
    def time(self, v):
        assert v is None or isinstance(v, (str, int)), 'Status.time can only be None, str or int'
        
        self._time = datetime.datetime.fromtimestamp(int(v))
    
    @property
    def percent(self):
        return self._percent
    
    @percent.setter
    def percent(self, v):
        assert v is None or isinstance(v, (str,float)), 'Status.percent can only be None, str or float'
        
        self._percent = float(v)
    
    @property
    def remaining(self):
        return self._remaining
    
    @remaining.setter
    def remaining(self, v):
        assert v is None or isinstance(v, (str, int)), 'Status.remaining can only be None, str or int'
        
        if v:
            self._remaining = int(v)
        else:
            self._remaining = None
    
    @property
    def etc(self):
        return self._etc
    
    @etc.setter
    def etc(self, v):
        assert v is None or isinstance(v, (str,int)), 'Status.etc can only be None, str or int'
        
        if v:
            self._etc = datetime.datetime.fromtimestamp(int(v))
        else:
            self._etc = None



#!/usr/bin/env python3

# Copyright (c) 2019 Christian Barral

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# This module contains the XML parser used to transform 
# pure Nmap XML output into Python objects.

import datetime

from typing import Any, Union, List, Tuple
from .exceptions import MissingScript


class Host:
    """ Holds the information from an individual target that responded to the scan.

    Such information is generic host information like the status, reason, reason TTL...
    Ports information, including the services objects, CPEs, and so on... and even hosts
    and services scripts. A Host instance may save other elements instances to hold information.
    """

    __slots__ = ('_state', '_reason', '_reason_ttl', '_start_time', '_end_time', '_ipv4', '_ipv6',
                 '_hostnames', '_ports', '_oses', '_fingerprint', '_trace', '_scripts', '_index')

    def __init__(self, **kwargs):
        self.state = kwargs.get('state', None)
        self.reason = kwargs.get('reason', None)
        self.reason_ttl = kwargs.get('reason_ttl', None)
        self.start_time = kwargs.get('start_time', None)
        self.end_time = kwargs.get('end_time', None)
        self.ipv4 = kwargs.get('ipv4', None)
        self.ipv6 = kwargs.get('ipv6', None)
        self.fingerprint = kwargs.get('fingerprint', None)
        self._hostnames = kwargs.get('hostnames', {})
        self._ports = kwargs.get('ports', [])
        self._oses = kwargs.get('oses', [])
        self._trace = kwargs.get('trace', [])
        self._scripts = kwargs.get('scripts', {})

        self._index = -1

    @property
    def state(self) -> Union[None,str]:
        """ Host state
        """
        return self._state
    
    @state.setter
    def state(self, v):
        assert v is None or isinstance(v, str), 'Host.state must be None or str'

        self._state = v

    @property
    def reason(self) -> Union[None,str]:
        """ Host's reason for a particular state
        """
        return self._reason

    @reason.setter
    def reason(self, v):
        assert v is None or isinstance(v, str), 'Host.reason must be None or str'

        self._reason = v

    @property
    def reason_ttl(self) -> Union[None,str]:
        """ Reason TTL from state retrieval
        """
        return self._reason_ttl

    @reason_ttl.setter
    def reason_ttl(self, v):
        assert v is None or isinstance(v, str), 'Host.reason_ttl must be None or str'

        self._reason_ttl = int(v)

    @property
    def start_time(self)  -> Union[None,datetime.datetime]:
        """ Datetime from when the host started to be scanned
        """
        return self._start_time

    @start_time.setter
    def start_time(self, v):

        if v is not None:
            self._start_time = datetime.datetime.fromtimestamp(int(v))
        else:
            self._start_time = None

    @property
    def end_time(self) -> Union[None,datetime.datetime]:
        """ Datetime from when the host stopped to be scanned
        """
        return self._start_time

    @end_time.setter
    def end_time(self, v):

        if v is not None:
            self._end_time = datetime.datetime.fromtimestamp(int(v))
        else:
            self._end_time = None

    @property
    def ip(self) -> Union[None,str]:
        """ Returns the host's Ipv4 if it has one associated, and Ipv6 in any other case
        """
        if self._ipv4 is not None:
            return self._ipv4
        else:
            return self._ipv6
    
    @property
    def ipv4(self) -> Union[None,str]:
        """ Returns the host's IPv4
        """
        return self._ipv4

    @ipv4.setter
    def ipv4(self, v):
        assert v is None or isinstance(v, str), 'Host.ipv4 must be None or str'

        self._ipv4 = v
    
    @property
    def ipv6(self) -> Union[None,str]:
        """ Returns the host's IPv6
        """
        return self._ipv6

    @ipv6.setter
    def ipv6(self, v):
        assert v is None or isinstance(v, str), 'Host.ipv6 must be None or str'

        self._ipv6 = v

    @property
    def fingerprint(self) -> Union[None,str]:
        """ Return the host's OS fingerprint
        """
        return self._fingerprint
    
    @fingerprint.setter
    def fingerprint(self, v):
        assert v is None or isinstance(v, str), 'Host.fingerprint must be None or str'

        self._fingerprint = v

    def __eq__(self, v):
        """ Returns True if this object is equaled to its IPv4, IPv6 or any of its hostnames
        
        :param v: Value to compare
        :returns: True if any IPv4, IPv6 or hostnames matches
        """
        return self.ipv4 == v or self.ipv6 == v or v in self.hostnames()

    def __len__(self):
        return len(self._ports)

    def __iter__(self):
        return iter(self._ports)
    
    def __next__(self):
        if (self._index + 1) < len(self._ports):
            self._index += 1
            return self._ports[self._index]
        else:
            raise StopIteration

    def _add_port(self, *args):
        """ Add a port object binded to the current instance

        :param args: Any number of Port instances
        """

        for i in args:
            if not isinstance(i, Port):
                raise TypeError('Cannot add non-Port instance to the host')
            
            self._ports.append(i)

    def _add_os(self, os):
        """ Bind an OperatingSystem object to the current instance

        :param os: OperatingSystem instance to buy.
        :raises TypeError: If os is not an OperatingSystem
        """

        if not isinstance(os, OperatingSystem):
            raise TypeError('Cannot bind non-OperatingSystem instance to host')
            
        self._oses.append(os)

    def _add_hops(self, *args):
        """ Add an arbitrary number of Hop instances to the current host trace information

        :raises TypeError if the paramer is not a Hop object
        """

        for hop_instance in args:
            if not isinstance(hop_instance, Hop):
                raise TypeError('Cannot bind a non-Hop object to a host`s trace')
            
            self._trace.append(hop_instance)

    def _add_script(self, script_name, script_output):
        """ Add a script name and output to the host scripts

        :param script_name: Name of the NSE script
        :param script_output: Output from the script execution
        """
        self._scripts[script_name] = script_output

    def scanned_ports(self):
        """ Returns the list of scanned ports
        
        :returns: List of scanned ports
        """
        return self._ports

    def udp_ports(self):
        """ Returns the list of scanned UDP ports
        
        :returns: List of scanned UDP ports
        """

        return [x for x in self._ports if x.protocol == 'udp']

    def tcp_ports(self):
        """ Returns the list of scanned TCP ports
        
        :returns: List of scanned TCP ports
        """

        return [x for x in self._ports if x.protocol == 'tcp']

    def hostnames(self, include_type: bool = False) -> Union[List[str],Tuple[str,str]]:
        """ Return all the host related hostnames.
        
        if include_type is set to True, the method will return a list of tuples where
        the first element is the hostname, and the second is the type of hostname related to the
        first element.

        :param include_type: Set to True to include the hostnames types on the result.
        :returns: List of hostnames or list of tuples with (hostname, hostname_type).
        """

        if not include_type:
            return [x for x in self._hostnames.keys()]
        else:
            return [(x, y) for x, y in self._hostnames.items()]

    def os_matches(self) -> list:
        """ Returns a list from all the OperatingSystem objects linked to the host
        
        :returns: List of operating systems
        """

        return self._oses

    def most_accurate_os(self):
        """ Returns the OperatingSystem object with the highest accuracy
        
        :returns: OperatingSystem or None if there were no OS matches
        """

        if not len(self._oses):
            return None
        else:
            return max(self._oses, key=lambda x: x.accuracy)

    def traceroute(self):
        """ Returns a list from all the Hop objects from a traceroute.
        
        :returns: List of Hops
        """

        return self._trace

    def get_script(self, script_name: str):
        """ Returns a script from host's scripts or raises MissingScript if it does not exist

        :param script_name: Name of the script
        :returns: Script output
        :raises: MissingScript if the given script is nor registered
        """
        if script_name in self._scripts:
            return self._scripts[script_name]
        else:
            raise MissingScript('No script output for the given script: {}'.format(script_name))

    def all_scripts(self):
        """ Returns a list of tuples containing the scripts names and outputs from all executed port scripts
        
        :returns: List of tuples
        """

        return [(x,y) for x,y in self._scripts.items()]

class Port:
    """ A port element represents a unique port from an individual protocol related to a host.

    A port element has information from the port number, the protocol, the port state and, if any,
    the service running on that port. Note that the service is not a primitive, but an instance of the
    Service class.
    """

    __slots__ = ('_protocol', '_number', '_state', '_reason', '_reason_ttl', '_service')

    def __init__(self, **kwargs):
        self.protocol = kwargs.get('protocol', None)
        self.number = kwargs.get('number', None)
        self.state = kwargs.get('state', None)
        self.reason = kwargs.get('reason', None)
        self.reason_ttl = kwargs.get('reason_ttl', None)
        self._service = None

    @property
    def protocol(self) -> Union[None,str]:
        """ Return the port transport protocol
        """
        return self._protocol

    @protocol.setter
    def protocol(self, v):
        assert v is None or isinstance(v, str), 'Port.protocol must be None or str'

        self._protocol = v

    @property
    def number(self)  -> Union[None,int]:
        """ Return the port number
        """
        return self._number
    
    @number.setter
    def number(self, v):
        assert v is None or isinstance(v, (str, int)) ,'Port.number must be None, str or int'

        self._number = int(v)

    @property
    def state(self) -> Union[None,str]:
        """ Return the port state, that can be 'open', 'filtered' or 'closed'"""
        return self._state

    @state.setter
    def state(self, v):
        assert v is None or isinstance(v, str), 'Port.state must be None or str'

        self._state = v

    @property
    def reason(self) -> Union[None,str]:
        """ Return the reason for a particular port state
        """
        return self._reason

    @reason.setter
    def reason(self, v):
        assert v is None or isinstance(v, str), 'Port.reason must be None or str'

        self._reason = v

    @property
    def reason_ttl(self) -> Union[None,int]:
        """ Return the reason's TTL
        """
        return self._reason_ttl

    @reason_ttl.setter
    def reason_ttl(self, v):
        assert v is None or isinstance(v, (str, int)), 'Port.reason_ttl must be None, str or int'

        self._reason_ttl = int(v)

    @property
    def service(self):
        return self._service

    def __eq__(self, v) -> bool:
        """ Evaluates the port instance against an integer, which directly compares self.port with 
        the value to be compared

        :returns: True if the ports are the same or False in any other case
        """

        if not isinstance(v, int):
            raise TypeError('Cannot compare Port with non-integer variables')
        
        return self.number == v

    def _add_service(self, service):
        """ Bind a service with the current instance
        
        :param service: Service instance to bind
        """

        if not isinstance(service, Service):
            raise TypeError('Cannot bind a non-Service instance to a port')

        self._service = service  

    def get_service(self):
        """ Returns a potential Service instance linked to the port, or None in any other case
        
        :returns: Service instance or None
        """

        return self._service


class Service:
    """ Represents a service binded to a port.

    A service can be associated with a name, product, version and extrainfo. It addtionally
    has information about the tunneling protocol, service CPE(s) and scripts.
    """

    __slots__ = ('_name', '_product', '_version', '_extrainfo', '_tunnel', '_method',
                 '_conf', '_cpes', '_scripts', '_port')

    def __init__(self, **kwargs):
        self.name = kwargs.get('name', None)
        self.product = kwargs.get('product', None)
        self.version = kwargs.get('version', None)
        self.extrainfo = kwargs.get('extrainfo', None)
        self.tunnel = kwargs.get('tunnel', None)
        self.method = kwargs.get('method', None)
        self.conf = kwargs.get('conf', None)
        self.cpes = kwargs.get('cpes', [])
        self.port = kwargs.get('port', None)
        
        self._scripts = {}

    @property
    def name(self) -> Union[None,str]:
        """ Return the service name
        """
        return self._name

    @name.setter
    def name(self, v):
        assert v is None or isinstance(v, str), 'Service.name must be None or str'

        self._name = v

    @property
    def product(self) -> Union[None,str]:
        """ Return the product being executed behind the service
        """
        return self._product

    @product.setter
    def product(self, v):
        assert v is None or isinstance(v, str), 'Service.product must be None or str'

        self._product = v

    @property
    def version(self) -> Union[None,str]:
        """ Return the product version
        """
        return self._version

    @version.setter
    def version(self, v):
        assert v is None or isinstance(v, str), 'Service.version must be None or str'

        self._version = v
    
    @property
    def extrainfo(self) -> Union[None,str]:
        """ Return trivial extra information about a particular service
        """
        return self._extrainfo

    @extrainfo.setter
    def extrainfo(self, v):
        assert v is None or isinstance(v, str), 'Service.extrainfo must be None or str'

        self._extrainfo = v

    @property
    def tunnel(self) -> Union[None,str]:
        """ Return the tunneling method used by the service
        """
        return self._tunnel

    @tunnel.setter
    def tunnel(self, v):
        assert v is None or isinstance(v, str), 'Service.tunnel must be None or str'

        self._tunnel = v
    
    @property
    def method(self) -> Union[None,str]:
        """ Return the method used to identify the service
        """
        return self._method

    @method.setter
    def method(self, v):
        assert v is None or isinstance(v, str), 'Service.method must be None or str'

        self._method = v
    
    @property
    def conf(self) -> Union[None,float]:
        """ Return the confidentiality from Nmap about the information reported for a particular service
        """
        return self._conf

    @conf.setter
    def conf(self, v):
        assert v is None or isinstance(v, str), 'Service.conf must be None or str'

        self._conf = float(v)

    @property
    def cpes(self) -> Union[None,list]:
        """ Return a list of all identified CPEs for a particular service
        """
        return self._cpes

    @cpes.setter
    def cpes(self, v):
        assert v is None or isinstance(v, list), 'Service.cpes must be None or list'

        self._cpes = v

    @property
    def port(self) -> Union[None,int]:
        """ Return the port number from where the service was identified
        """
        return self._port

    @port.setter
    def port(self, v):
        assert v is None or isinstance(v, (str, int)), 'Service.port must be None, str or int'

        self._port = int(v)

    def _add_script(self, script_name, script_output):
        """ Add a script name + output to tthe service instance.

        :param script_name: Script identifier
        :param script_output: Script output
        """

        if script_name in self._scripts:
            raise RuntimeError('Script with identifier "{}" already exists, please name it differently'.format(script_name))

        self._scripts[script_name] = script_output

    def get_script(self, script_name: str):
        """ Returns a port script associated with this service.
        
        :param script_name: Name of the script to return
        :returns: Script output
        :raises: MissingScript if the script_name is not on the instance's scripts
        """

        if script_name in self._scripts:
            return self._scripts[script_name]
        else:
            raise MissingScript('No script output for the given script: {}'.format(script_name))

    def all_scripts(self):
        """ Returns a list of tuples containing the scripts names and outputs from the host
        
        :returns: List of tuples
        """

        return [(x,y) for x,y in self._scripts.items()]

class OperatingSystemMatch:
    """ Represents a single match from an operating system.

    It contains information from the type, vendor, family, generation and CPE. An operating
    system match is itself part of an OperatingSystem. When Nmap identifies the host's OS, it 
    outputs its generic information (name and accuracy), but the findings that made Nmap decide
    the host's OS are 1-N relation, so this class represents each of those N findings.
    """

    __slots__ = ('_type', '_vendor', '_family', '_generation', '_cpe')

    def __init__(self, **kwargs):
        self.type = kwargs.get('type', None)
        self.vendor = kwargs.get('vendor', None)
        self.family = kwargs.get('family', None)
        self.generation = kwargs.get('generation', None)
        self.cpe = kwargs.get('cpe', None)

    @property
    def type(self):
        """ Return the OS type
        """
        return self._type
    
    @type.setter
    def type(self, v):

        assert v is None or isinstance(v, str), 'OperatingSystemMatch.type must be an str or None'

        self._type = v
    
    @property
    def vendor(self):
        """ Return the OS vendor
        """
        return self._vendor
    
    @vendor.setter
    def vendor(self, v):

        assert v is None or isinstance(v, str), 'OperatingSystemMatch.vendor must be an str or None'

        self._vendor = v
    
    @property
    def family(self):
        """ Return the OS family
        """
        return self._family
    
    @family.setter
    def family(self, v):

        assert v is None or isinstance(v, str), 'OperatingSystemMatch.family must be an str or None'

        self._family = v

    @property
    def generation(self):
        """ Return the OS generation
        """
        return self._generation
    
    @generation.setter
    def generation(self, v):

        assert v is None or isinstance(v, str), 'OperatingSystemMatch.generation must be None or str'

        self._generation = v
    
    @property
    def cpe(self):
        """ Return the OS CPE
        """
        return self._cpe

    @cpe.setter
    def cpe(self, v):

        assert v is None or isinstance(v, str), 'OperatingSystemMatch.cpe must be an str or None'

        self._cpe = v


class OperatingSystem:
    """ Represents a host's operating system scan

    An operating system contains information from an OS match, which includes the OS name and accuracy. 
    An operating system will also have one or more OS matches, which will hold the information of why an operating 
    system matched with the target.
    """

    __slots__ = ('_name', '_accuracy', '_matches')

    def __init__(self, **kwargs):
        self.name = kwargs.get('name', None)
        self.accuracy = kwargs.get('accuracy', None)
        self._matches = []

        # Add all matches objects
        for match_info in kwargs.get('matches', []):
            self._matches.append(OperatingSystemMatch(**match_info))

    @property
    def name(self) -> Union[None,str]:
        """ Return a string representing the operating system name and version
        """
        return self._name

    @name.setter
    def name(self, v):
        assert v is None or isinstance(v, str), 'OperatingSystem.name must be None or str'

        self._name = v

    @property
    def accuracy(self) -> Union[None,float]:
        """ Return the accuracy between 0 and 100 related to the OS match
        """
        return self._accuracy

    @accuracy.setter
    def accuracy(self, v):
        assert v is None or isinstance(v, str), 'OperatingSystem.accuracy must be None or str'

        self._accuracy = float(v)

    def get_matches(self) -> List[OperatingSystemMatch]:
        """ Return all single matches made by Nmap, which lead to the current overall Operating System matching
        
        :returns: List of objects representing each single match
        """
        return self._matches


class Hop:
    """ A Hop is a representantion of an ICMP hop performed by traceroute.

    It saves information from the hostname, IP, Rount-Trip-Time and Time-To-Live
    """

    __slots__ = ('_host', '_ip', '_rtt', '_ttl')

    def __init__(self, **kwargs):
        self.host = kwargs.get('host', None)
        self.ip = kwargs.get('ip', None)
        self.rtt = kwargs.get('rtt', None)
        self.ttl = kwargs.get('ttl', None)

    @property
    def host(self):
        """ Return the noce hostname
        """
        return self._host
    
    @host.setter
    def host(self, v):
        assert v is None or isinstance(v, str), 'Hop.host must None or str'

        self._host = v
    
    @property
    def ip(self):
        """ Return the node IP address
        """
        return self._ip
    
    @ip.setter
    def ip(self, v):
        assert v is None or isinstance(v, str), 'Hop.ip must None or str'

        self._ip = v
    
    @property
    def rtt(self):
        """ Return the Round-Trip-Time
        """
        return self._rtt
    
    @rtt.setter
    def rtt(self, v):
        assert v is None or isinstance(v, str), 'Hop.rtt must None or str'

        self._rtt = v
    
    @property
    def ttl(self):
        """ Return the Time-To-Live
        """
        return self._ttl
    
    @ttl.setter
    def ttl(self, v):
        assert v is None or isinstance(v, str), 'Hop.ttl must None or str'

        self._ttl = v



#!/usr/bin/env python

# Copyright (c) 2019 Christian Barral

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ['NmapScanError', 'InvalidPortError', 'InvalidArgumentError', 'MalformedIpAddressError',
           'EngineError', 'XMLParsingError']


class NmapScanError(Exception):
    """ Exception class for nmap scanning errors.
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class InvalidPortError(NmapScanError):
    """ Exception class for port assignment and parsing errors.
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class MalformedIpAddressError(NmapScanError):
    """ Exception class for target assignment and parsing errors.
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class InvalidArgumentError(NmapScanError):
    """ Exception class for nmap arguments assignment and parsing errors.
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class XMLParsingError(Exception):
    """ Exception class for nmap output parsing errors.
    """

    def __init__(self, message):
        Exception.__init__(self, message)

class MissingScript(Exception):
    """ Exception to be raised when a script is missing
    """

    def __init__(self, message):
        super().__init__(message)


class EngineError(Exception):
    """ Exception class for PyNSEEngine errors
    """

    def __init__(self, msg):
        super().__init__(msg)

class StopExecution(Exception):
    """ Exception for knowing when to not add script output
    """

    def __init__(self, msg):
        super().__init__(msg)


#!/usr/bin/env python3

# Copyright (c) 2019 Christian Barral

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# This module contains the XML parser used to transform 
# pure Nmap XML output into Python objects.


from collections.abc import Iterable

from . import utils
from .exceptions import InvalidArgumentError, InvalidPortError


class _PortAbstraction:
    """ Encapsulates the logic for using tcp(), udp() and top_ports() functions together, and
    generates the different syntax to be used for Nmap depending on the 
    selected options.

    A user may not select TCP or UDP ports twice, and a user may not either select top_ports with other TCP or UDP ports.
    """

    def __init__(self):
        self._tcp_ports = None
        self._udp_ports = None
        self._top_ports = None
        self._malleable_ports = None

        # Flags
        self._has_added_tcp = False
        self._has_added_udp = False
        self._has_added_top_ports = False
        self._has_added_malleable = False

    @staticmethod
    def _parse_port_range(port_range):
        """ Parses a port range, depending on if its an str or an Iterable.

        :param port_range: Port range to parse
        :returns: String representing the port range.
        """

        # If int
        if isinstance(port_range, int):
            if not utils.valid_port(port_range):
                raise InvalidPortError('Invalid port number: {}'.format(port_range))
            
            return str(port_range)

        # If string
        elif isinstance(port_range, str):
            port_range = port_range.strip()

            # If port range is a word
            if port_range == 'all' or port_range == '*':
                return '-'
            
            # In any other case, suppose that it is an str port range
            # Execute ports to list to raise any possible InvalidPortError
            _ = utils.ports_to_list(port_range)
            return port_range
        
        # If iterable
        elif isinstance(port_range, Iterable):

            return utils.ports_to_str(port_range)
        
        # In any other case, raise error
        else:
            raise TypeError('Invalid port_range type: {}, expected str or Iterable'.format(type(port_range)))

    def tcp(self, port_range):
        """ Adds a port range to the TCP port selection.
        """

        if self._has_added_tcp:
            raise InvalidPortError('Cannot add TCP ports twice.')
        
        if self._has_added_top_ports:
            raise InvalidPortError('Cannot specify top ports and individual ports on a same scanner.')

        if port_range:
            self._tcp_ports = self._parse_port_range(port_range)
        
        self._has_added_tcp = True
        return self
    
    def udp(self, port_range):
        """ Adds a port range to the UDP port selection.
        """

        if self._has_added_udp:
            raise InvalidPortError('Cannot add UDP ports twice.')
        
        if self._has_added_top_ports:
            raise InvalidPortError('Cannot specify top ports and individual ports on a same scanner.')
        
        if port_range:
            self._udp_ports = self._parse_port_range(port_range)
        
        self._has_added_udp = True
        return self
    
    def top_ports(self, num_ports):
        """ Adds a --top-ports nmap parameter with num_ports
        
        :param num_ports: Number of ports to add to the scanner.
        :returns: Instance of the object itself
        """

        if self._has_added_tcp or self._has_added_udp:
            raise InvalidPortError('Cannot specify UDP and TCP when also specifying top ports.')

        try:
            num_ports = int(num_ports)
        except ValueError:
            raise InvalidPortError('Invalid top_ports value, expected int but got {}'.format(type(num_ports)))
        
        if not 1 <= num_ports <= 65535:
            raise InvalidPortError('Invalid top_ports value, must be between 1 and 65535')
        
        self._top_ports = num_ports
        self._has_added_top_ports = True

        return self

    def _malleable(self, port_range):
        """ Adds a malleable port range. 
        
        Malleable ports are those that are protocol-specific. This means that a user may specify ports without the tcp() or udp() functions/methods.
        so the protocol to be used with those ports depends on the user specifying -sU, any other option or -sU + any other option. In this case, malleable ports
        should not be added to the raw nmap command with T: or U:
        
        :param port_range: Port range to add.
        :returns: Instance of the object itself
        """

        # Since this is an internal function that should NOT be used directly, no other checks are needed
        if port_range:
            self._malleable_ports = self._parse_port_range(port_range)
        
        self._has_added_malleable = True
        return self
        
    
    def to_nmap_syntax(self):
        """ Based on the object data, it returns a valid nmap string that can be directly injected into an nmap command.

        :returns: String with nmap-like syntax
        """

        if self._has_added_malleable:
            return self._malleable_ports

        elif self._top_ports:
            return '--top-ports {}'.format(self._top_ports)

        elif self._tcp_ports or self._udp_ports:
            nmap_port_string = ''
            if self._tcp_ports:
                nmap_port_string += 'T:{}'.format(self._tcp_ports)
            if self._udp_ports and self._tcp_ports:
                nmap_port_string += ',U:{}'.format(self._udp_ports)
            elif self._udp_ports:
                nmap_port_string += 'U:{}'.format(self._udp_ports)
            
            return nmap_port_string
        
        else:
            return None


def tcp(port_range):
    """ Returns a TCP port range encapsulated into a PortAbstraction object

    :param port_range: String or Iterable of ports to set to the Nmap command.
    :returns: Instance of _PortAbstraction with the added port_range as TCP ports
    """
    
    return _PortAbstraction().tcp(port_range)


def udp(port_range):
    """ Returns a UDP port range encapsulated into a PortAbstraction object

    :param port_range: String or Iterable of ports to set to the Nmap command.
    :returns: Instance of _PortAbstraction with the added port_range as UDP ports
    """
    
    return _PortAbstraction().udp(port_range)


def top_ports(num_ports):
    """ Returns Nmap-like syntax for top ports

    :param num_ports: Number of ports
    """

    return _PortAbstraction().top_ports(num_ports)


#!/usr/bin/env python3

# Copyright (c) 2019 Christian Barral

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# This module defines async operations for Nmap scans

import pathlib
import random
import os
import subprocess
import string

from typing import Union, Iterable, Tuple

from .scanner import NmapScanner
from .results import NmapScanResult
from .exceptions import NmapScanError, XMLParsingError
from .engine import NSE
from .ports import _PortAbstraction
from .status import Status


class NmapAsyncScanner(NmapScanner):
    """ Allows async operations for Nmap scans, like background execution, status retrieval and process awaiting
    
    It works like the NmapScanner object but has some additional methods and kwargs on its methods. The class overrides some of the core 
    NmapScanner functionality in order allow the manipulation of the Nmap's child process.
    """

    def __init__(self, nmap_bin: Union[None,str] = None, engine: Union[None,NSE] = None):
        super().__init__(nmap_bin, engine)
        self._nmap_process = None
        self._has_started = False
        self._has_awaited = False

        self._stored_output_buffer = None
        self._stored_error_buffer = None

        # Flags to check execution type
        self._requires_file_parsing = False
        self._xml_file_path = None

        # Save any needed information to be used across multiple methods
        self._priority_engine = None
        self._output_base_filename = None

        # Status flags
        self._with_status = False
        self._retrieved_status = False
        self._last_status_instance = None

    @property
    def _selected_engine(self):
        if self._priority_engine:
            return self._priority_engine
        else:
            return self._engine

    def _add_stats_every(self, command, interval) -> str:
        """ Extends a Nmap command with the --stats-every agument
        
        :param command: Command to extend
        :param interval: Stats every interval
        """
        nmap_command_length = 4 if not self._nmap_bin else len(self._nmap_bin)

        return fr'{command[:nmap_command_length]} --stats-every {interval}{command[nmap_command_length:]}'


    def _execute_nmap(self, nmap_arguments) -> Tuple[bytes,bytes]:
        """ Execute and asynchronous Nmap process

        :param nmap_arguments: List of Nmap arguments
        :raises NmapScanError: If the provided Nmap binary path is not valid.
        """

        # Popen raises FileNotFoundError in case the program does not exist
        try:
            self._has_started = True
            self._nmap_process = subprocess.Popen(nmap_arguments, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except FileNotFoundError:
            raise NmapScanError('Nmap was not found on the system. Please install it before using Nmapthon2') from None

    def scan(self, targets: Union[str,Iterable], ports: Union[None,int,str,Iterable,_PortAbstraction] = None,  arguments: Union[None,str] = None, 
             dry_run: bool = False, output: Union[None,str,Iterable] = None, engine: Union[None,NSE] = None, with_status: bool = False, status_interval: str = '3s'):
        """ Execute an Nmap scan based on on a series of targets, and optional ports and
        arguments. For multi-output format storage the output argument can be set with 
        the needed extersions or output parameters. It also accepts an NSE engine that will override the instance`s engine, in case there is one.
        It works with asynchronous operations.

        :param targets: Targets to scan inside a str or Iterable type, like a list. Targets can also be specified through network ranges, partial ranges, network with CIDR mask and domains/hostnames.
        :param ports: Ports to scan in as an int, str, iterable or custom functions. Ports can also be specified with ranges.
        :param arguments: Arguments to execute Nmap in a single string
        :param dry_run: Set to True if you just want to test your parameters, with this option the scan does NOT run.     
        :param output: Iterable with desired output formats, that can be "xml", "normal" and/or "grep".
        :param engine: NSE object for custom script execution. It overrides the NSE object specified on the instance for the current scan.
        :param with_status: If set to True, it appends status information to the Nmap output every status_interval. Defalt: False
        :param status_interval: Status interval to STDOUT in Nmap-like format. Default: '5s'
        :raises NmapScanError: If you call scan() on an instance that is already scanning
        """

        if self._has_started:
            raise NmapScanError('Cannot call scan() on an instance that is already scanning')

        # Set priority engine and started flag
        self._has_started = True
        if engine:
            self._priority_engine = engine

        random_nmap_output_filename = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(25))
        nmap_command = self._create_nmap_command(targets, random_nmap_output_filename, ports, arguments, output)

        # If dry_run, do not execute
        if dry_run:
            return

        # If with_status, add to command
        if with_status:
            self._with_status = True
            self._stored_output_buffer = ''
            nmap_command = self._add_stats_every(nmap_command, status_interval)

        # Store output base filename if output was specified, and say that it needs file processing
        if output:
            self._requires_file_parsing = True
            self._xml_file_path = os.path.join(self._temp_folder, f'{random_nmap_output_filename}.xml')
            self._output_base_filename = random_nmap_output_filename

        self._execute_nmap(nmap_command)

    def raw(self, raw_arguments: str, engine: Union[None,NSE] = None):
        """ Executes a Nmap scan with a raw string containing all the command itself, without the 'nmap' keyword, but with asynchronous opeartions

        :param raw_arguments: String containing all nmap arguments. No --resume option nor output options should be placed here.
        :param engine: NSE instance to use for this particular scan.
        :raises NmapScanError: If raw() is called on an instance that is already scanning
        """

        if self._has_started:
            raise NmapScanError('Cannot call raw() on a instance that is already scanning')

        # Set priority engine and started flag
        self._has_started = True
        if engine:
            self._priority_engine = engine

        if raw_arguments.startswith('nmap '):
            raw_arguments = raw_arguments[5:]
        
        if any([x in raw_arguments for x in ('--resume', '-oA', '-oX', '-oN', '-oS', '-oG')]):
            raise NmapScanError('Cannot specify --resume nor output options.')
        
        if self._nmap_bin:
            nmap_bin = self._nmap_bin
        else:
            nmap_bin = 'nmap'
        
        raw_arguments = '{} {} -oX -'.format(nmap_bin, raw_arguments)
        
        self._execute_nmap(self._split_command(raw_arguments))

    def resume(self, xml_file: Union[pathlib.Path,str]):
        """ Resumes an Nmap scan from an XML file, but with asynchronous operations

        :param xml_file: String representing the file's path
        :raises NmapScanError: If you call resume on an instance that has already started a scan
        """
        if self._has_started:
            raise NmapScanError('Cannot call resume() on an instance that is already scanning')
        
        # Set the resume and has_started flag
        self._requires_file_parsing = True
        self._has_started = True

        if isinstance(xml_file, pathlib.Path):
            xml_file = xml_file.absolute

        # Set to instance attribute for later processing
        self._xml_file_path = xml_file

        if self._nmap_bin:
            nmap_bin = self._nmap_bin
        else:
            nmap_bin = 'nmap'

        self._execute_nmap(self._split_command('{} --resume {}'.format(nmap_bin, xml_file)))
            
    def get_result(self) -> Union[None,NmapScanResult]:
        """ Returns the NmapScanResult object from the performed scan and raises any non-tolerant errors return by Nmap, if any.
        
        :raises NmapScanError: When trying to get hte result from a scan that has not started
        :returns: Scan result
        """

        if not self._has_started:
            raise NmapScanError('Cannot get a result from a scan that has not started')

        # If blocking, just let the function continue, since the communicate() call will freeze the thread
        if not self.finished():
            return None
        
        # Check what buffers to use, the already stored ones or the processes pipes
        # Also, check if status has been retrieved, since buffers may need to be concatenated, as the operating system flushes
        # the processes stdout every team we read a 256 chunk.
        if self._has_awaited:
            output_buff, err_buff = self._stored_output_buffer, self._stored_error_buffer
        elif self._retrieved_status:
            output_buff = self._stored_output_buffer + self._nmap_process.stdout.read().decode('latin-1')
            err_buff = self._nmap_process.stderr.read().decode('latin-1')
        else:
            output_buff, err_buff = self._nmap_process.communicate()

        try:
            # If resume(), parse the stored XML file in the instance attribute
            if self._requires_file_parsing:
                if not err_buff:
                    with open(self._xml_file_path) as f:
                        return self._parse_nmap_output(output_buff, err_buff, output=self._output_base_filename, engine=self._priority_engine)
                else:
                    raise NmapScanError(err_buff.decode('utf8'))
            else:
                return self._parse_nmap_output(output_buff, err_buff, output=self._output_base_filename, engine=self._priority_engine)
        finally:
            # Reset values for the next scan
            if self._output_base_filename:
                self._delete_output_files(self._output_base_filename)

    def finished(self) -> bool:
        """ Returns True or False depending on whether the Nmap process has ended or not
        
        :raises NmapScanError: If no scan has been started
        """
        if not self._has_started:
            raise NmapScanError('Cannot check if a scan has finished without starting it')

        return self._nmap_process.poll() is not None
    
    def wait(self):
        """ Freezes the main process until Nmap finishes

        :raises NmapScanError: If no scan has been started
        """
        if not self._has_started:
            raise NmapScanError('Cannot wait on a scan that has not started')

        # Set the has_awaited flag to True, which will cause the get_result method to return the result based on the already stored buffers
        self._has_awaited = True
        temp_output_buffer, self._stored_error_buffer = self._nmap_process.communicate()

        if self._retrieved_status:
            self._stored_output_buffer += temp_output_buffer.decode('latin-1')
        else:
            self._stored_output_buffer = temp_output_buffer

    def get_status(self) -> Union[None,Status]:
        """ Return a Status object representing Nmap's status. It returns None if no status could be parsed yet.

        :returns: Status representing Nmap's status
        :raises NmapScanError: If get_status() is called without previously setting with_status=True in the scan() method
        """
        if not self._with_status:
            raise NmapScanError('Cannot get Nmap status without setting with_status.')
        
        self._retrieved_status = True

        # If Nmap is writing its output to a file, then read the file instead of reading the process buffer
        # Catch FileNotFoundError to control race conditions with Nmap's child process
        if self._requires_file_parsing:
            try:
                with open(self._xml_file_path) as f:
                    return Status.from_raw_xml(f.read())
            except FileNotFoundError:
                return None

        # In any other case, just append to buffer
        else:
            line = self._nmap_process.stdout.read(256).decode('latin-1')
            if line:
                self._stored_output_buffer += line
            
            return Status.from_raw_xml(self._stored_output_buffer)

    def reset(self):
        """ Reset all the flags and scan-related variables to its original value
        """
        self._output_base_filename = None
        self._priority_engine = None
        self._stored_output_buffer = None
        self._stored_error_buffer = None
        self._has_started = False
        self._has_awaited = False
        self._requires_file_parsing = False
        self._xml_file_path = None
        self._with_status = False
        self._retrieved_status = False
        self._last_status_instance = None


#!/usr/bin/env python3

# Copyright (c) 2019 Christian Barral

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# This module contains the XML parser used to transform 
# pure Nmap XML output into Python objects.

import pathlib
import shlex
import tempfile
import subprocess
import random
import string
import os

from collections.abc import Iterable
from typing import Coroutine, Tuple, Union
import nmapthon2

from nmapthon2.results import NmapScanResult

from .parser import XMLParser
from .exceptions import InvalidArgumentError, NmapScanError, XMLParsingError
from .ports import _PortAbstraction
from .engine import NSE


OUTPUT_FORMATS = ('all', 'xml', 'normal', 'grep')
OUTPUT_RELATION = {
    'xml': '.xml',
    'normal': '.nmap',
    'grep': '.gnmap'
}


class NmapScanner:
    """ Represents a reusable Nmap Network scanner that wraps the results into
    NmapScanResult objects.

    It offers several methods to execute Nmap scans by different means, which are: 
    1) Specify targets, ports and arguments in three different arguments. 2) Specify the raw Nmap command. 
    3) Resume an Nmap scan

    Nmap does not allow multi format output to STDOUT, so this class may write to 
    the operating system temporal folder to write all the output formats, process them, and 
    delete them.

    :param nmap_bin: Path to the binary Nmap file.
    :param engine: Default NSE object for all scans performed with this scanner.
    """

    def __init__(self, nmap_bin: Union[None,str] = None, engine: Union[None,NSE] = None):

        self._temp_folder = tempfile.gettempdir()
        self._xml_parser = XMLParser()

        assert nmap_bin is None or isinstance(nmap_bin, str), 'nmap_bin must be None or str'
        assert engine is None or isinstance(engine, NSE), 'engine must be None or an instance of NSE'
        self._nmap_bin = nmap_bin
        self._engine = engine

    @staticmethod
    def _split_command(command: str) -> Union[list,str]:
        """ Split a command into a list of strings in UNIX systems, but leave the command as a single string for Windows systems.
        
        Commands including paths with Windows delimeters (\) will be missprocessed, causing errors in Windows systems. Nonetheless, 
        Windows systems do not require an array of arguments like UNIX, so passing the raw command is completely fine.

        :param command: Command to process
        """

        if os.name == 'nt':
            return command
        else:
            return shlex.split(command)

    def _parse_command_line_arguments(self, arguments_string):
        """ Parse the command line arguments from a given arguments string.

        Arguments have a few restrictions, which are: 1) the "nmap" command itself
        is added by Nmapthon2, so the arguments command must not start with it. 2) No
        output command should be passed (-oA, -oN, -oG, -oX or -oS), since by default 
        -oX - will be used, and in case user needs multi-output it is needed to use
        the output kwarg through the scan() method. 3) No --resume option should be used,
        the resume() method must be called instead.

        :param arguments_string: Raw arguments string.
        """

        # Split into a list of commands
        split_arguments = self._split_command(arguments_string)

        if '--resume' in split_arguments:
            raise NmapScanError('Cannot use --resume as a Nmap argument. Use resume() instead')

        if '-oX' in split_arguments or '-oN' in split_arguments or \
            '-oA' in split_arguments or '-oG' in split_arguments or '-oS' in split_arguments:
            raise NmapScanError('Cannot especify an output argument. Use the "output" kwarg instead.')

        if isinstance(split_arguments, list):
            return ' '.join(split_arguments)
        else:
            return split_arguments
    
    @staticmethod
    def _parse_output_flag(output):
        """ Parses the output wkarg from the scan() method.
        
        :returns: Iterable with output options correctly parsed.
        """

        # If str
        if isinstance(output, str):
            if output == 'kiddie':
                raise InvalidArgumentError('You should not be using this library, young padawan.')

            if output not in OUTPUT_FORMATS:
                raise InvalidArgumentError('Scan output must be of a valid type: {}'.format(', '.join(OUTPUT_FORMATS)))
                
            # Set the output to all other options if "all" is specified
            if output == 'all':
                return OUTPUT_FORMATS[1:]
            else:
                return (output,)

        # If iterable, validate them and change the variable to each type if "all" is specified
        elif isinstance(output, Iterable):
            if 'kiddie' in output:
                raise InvalidArgumentError('You should not be using this library, young padawan.')

            if 'all' in output:
                return OUTPUT_FORMATS[1:]

            else:
                for i in output:
                    if i not in OUTPUT_FORMATS:
                        raise InvalidArgumentError('Invalid output value: {}'.format(i))
            
            return output

        # Raise error in any other case
        else:
            raise TypeError('output parameter must be a string or an iterable with valid format types')

    @staticmethod
    def _parse_ports_flag(ports):
        """ Parses the ports flag from the scan() method
        
        :returns: Parsed string
        """

        if isinstance(ports, (str, int)) or isinstance(ports, Iterable):
            return _PortAbstraction()._malleable(ports).to_nmap_syntax()
        elif isinstance(ports, _PortAbstraction):
            return ports.to_nmap_syntax()
        else:
            raise InvalidArgumentError('Invalid type for ports. Expecting str, Iterable or specific function calling, but got: {}'.format(type(ports)))

    def _parse_targets(self, targets):
        """ Parses the targets from the scan() method
        
        :returns: Parsed targets
        """

        if isinstance(targets, str):
            if ' nmapthon ' in targets:
                raise NmapScanError('You dare to scan me?')
            _ = self._split_command(targets)
            return targets
        elif isinstance(targets, Iterable):
            if 'nmapthon' in targets:
                raise NmapScanError('You dare to scan me?')
            targets_str = ' '.join(targets)
            _ = self._split_command(targets_str)
            return targets_str
        else:
            raise InvalidArgumentError('Invalid targets type, expected str or Iterable, but got {}'.format(type(targets)))

    def _delete_output_files(self, random_nmap_output_filename):
        """ Deletes all generated files from Nmap
        
        :param random_nmap_output_filename: Random string to be used for file generation
        """
        for i in ('.xml', '.gnmap', '.nmap'):
            try:
                os.remove(os.path.join(self._temp_folder, random_nmap_output_filename, i))
            except FileNotFoundError:
                pass

    def _execute_nmap(self, nmap_arguments) -> Tuple[bytes,bytes]:
        """ Execute nmap and return the STDOUT and STDERR from the child process created

        :param nmap_arguments: List of Nmap arguments
        :raises NmapScanError: If the provided Nmap binary path is not valid.
        """

        # Popen raises FileNotFoundError in case the program does not exist
        try:
            nmap_process = subprocess.Popen(nmap_arguments, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except FileNotFoundError:
            raise NmapScanError('Nmap was not found on the system. Please install it before using Nmapthon2') from None

        return nmap_process.communicate()

    def _create_nmap_command(self, targets, random_nmap_base_filename, ports, arguments, output) -> list:
        """ Execute an Nmap scan based on on a series of targets, and optional ports and
        arguments. For multi-output format storage the output argument can be set with 
        the needed extersions or output parameters.

        :param targets: List of targets in an Iterable or str.
        :param random_nmap_base_filename: Random Nmap filename to create output files
        :param ports: Ports in str or list format
        :param arguments: Arguments to execute within the scan.
        :param output: Tuple or list of output formats.
        """

        # Validate, parse parameters and add them to the command
        
        # Target parsing
        targets = self._parse_targets(targets)

        # Add the commands
        if self._nmap_bin:
            nmap_command = self._nmap_bin
        else:
            nmap_command = 'nmap '

        # Ports
        if ports:
            ports = self._parse_ports_flag(ports)
            if '--top-ports' in ports:
                nmap_command += '{} '.format(ports)
            else:
                nmap_command += '-p{} '.format(ports)
        
        # Arguments
        if arguments:
            arguments = self._parse_command_line_arguments(arguments)
            nmap_command += '{} '.format(arguments)
        
        # Depending on the output argument, should add '-oX -' or start handling output through temp files.
        if output:
            output = self._parse_output_flag(output)
            nmap_command += '-oA {} '.format(os.path.join(self._temp_folder, random_nmap_base_filename))
        else:
            nmap_command += '-oX - '

        nmap_command += targets

        return self._split_command(nmap_command)

    def _parse_nmap_output(self, exec_output, exec_error, output: Union[None,str] = None, engine: Union[None,NSE] = None, skip_processing: bool = False) -> NmapScanResult:
        """ Parses the Nmap output comming from its execution through the child process, performs any required validations 
        and cleans the filesystem in case any files were created.

        :param exec_output: Child process' STDOUT
        :param exec_error: Child process' STDERR
        :param output: Single value representing the random filename assigned to the created files by the output flag
        :param engine: NSE object to execute after the scan
        :param skip_processing: Set to true, it does not try to process Nmap output. Should be used when resumming.
        """
        if len(exec_output):
            
            # If no output was set, parse directly from output
            if output is None:
                try:
                    result = self._xml_parser.parse_plain(exec_output)
                except XMLParsingError as e:
                    raise NmapScanError(exec_error.decode('utf8'))
            
                if skip_processing:
                    return None

            # If output was set, parse it from XML output file.
            else:
                try:
                    result = self._xml_parser.parse_file(os.path.join(self._temp_folder, '{}.xml'.format(output)))
                except XMLParsingError:
                    raise NmapScanError(exec_error.decode('utf8'))
                
                outputs = { 'xml': None, 'normal': None, 'grep': None }
                for i in outputs:
                    with open(os.path.join(self._temp_folder, '{}{}'.format(output, OUTPUT_RELATION[i]))) as f:
                        outputs[i] = f.read()
                
                result._normal_output = outputs['normal']
                result._grep_output = outputs['grep']
                result._xml_output = outputs['xml']

            # If execution reaches this point, then Nmapthon2 has parsed the XML correctly, but they might be tolerant errors remaining
            if len(exec_error):
                result.tolerant_errors = exec_error.decode('utf8')

            # Reference the coorect engine instance
            if not engine:
                engine = self._engine

            # Apply the engine parser and scripts execution to the Host, Port and Service instances
            if engine:
                for host in result:
                    # Apply any host script to the host object by reference
                    engine._apply_host_scripts(host)
                    for port in host:
                        service = port.get_service()
                        # If any parser to be used and there is a service with optential scripts, rock'em
                        if len(engine._parsers) and service:
                            for script_name, callback in engine._parsers.items():
                                try:
                                    service._scripts[script_name] = callback(service._scripts[script_name])
                                except KeyError as e:
                                    # If the KeyError is because of the script key not being in _scripts, then thats ok
                                    # but if not, should raise the exception to let know the programmer.
                                    if "'{}'".format(script_name) == str(e):
                                        pass
                                    else:
                                        raise
                        
                        # If any port script, apply it
                        engine._apply_port_scripts(host, port, service)

            return result
        
        else:
            if not skip_processing:
                raise NmapScanError(exec_error.decode('utf8'))
    
    def scan(self, targets: Union[str,Iterable], ports: Union[None,int,str,Iterable,_PortAbstraction] = None,  arguments: Union[None,str] = None, 
             dry_run: bool = False, output: Union[None,str,Iterable] = None, engine: Union[None,NSE] = None) -> NmapScanResult:
        """ Execute an Nmap scan based on on a series of targets, and optional ports and
        arguments. For multi-output format storage the output argument can be set with 
        the needed extersions or output parameters. It also accepts an NSE engine that will override the instance`s engine, in case there is one.

        :param targets: Targets to scan inside a str or Iterable type, like a list. Targets can also be specified through network ranges, partial ranges, network with CIDR mask and domains/hostnames.
        :param ports: Ports to scan in as an int, str, iterable or custom functions. Ports can also be specified with ranges.
        :param arguments: Arguments to execute Nmap in a single string
        :param dry_run: Set to True if you just want to test your parameters, with this option the scan does NOT run.     
        :param output: Iterable with desired output formats, that can be "xml", "normal" and/or "grep".
        :param engine: NSE object for custom script execution. It overrides the NSE object specified on the instance for the current scan.
        """

        if output:
            random_nmap_output_filename = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(25))
        else:
            random_nmap_output_filename = None

        nmap_command = self._create_nmap_command(targets, random_nmap_output_filename, ports, arguments, output)

        # If dry_run, do not execute
        if dry_run:
            return None

        try:
            output_buff, error_buff = self._execute_nmap(nmap_command)
            return self._parse_nmap_output(output_buff, error_buff, output=random_nmap_output_filename, engine=engine)
        finally:
            if output:
                self._delete_output_files(random_nmap_output_filename)
        

    def raw(self, raw_arguments: str, engine: Union[None,NSE] = None) -> NmapScanResult:
        """ Executes a Nmap scan with a raw string containing all the command itself, without the 'nmap' keyword.

        :param raw_arguments: String containing all nmap arguments. No --resume option nor output options should be placed here.
        :param engine: NSE instance to use for this particular scan.
        """

        if raw_arguments.startswith('nmap '):
            raw_arguments = raw_arguments[5:]
        
        if any([x in raw_arguments for x in ('--resume', '-oA', '-oX', '-oN', '-oS', '-oG')]):
            raise NmapScanError('Cannot specify --resume nor output options.')
        
        if self._nmap_bin:
            nmap_bin = self._nmap_bin
        else:
            nmap_bin = 'nmap'
        
        raw_arguments = '{} {} -oX -'.format(nmap_bin, raw_arguments)
        
        output_buff, error_buff = self._execute_nmap(self._split_command(raw_arguments))
        return self._parse_nmap_output(output_buff, error_buff, engine=engine)

    def resume(self, xml_file: Union[pathlib.Path,str]) -> NmapScanResult:
        """ Resumes an Nmap scan from an XML file.

        :param xml_file: String representing the file's path
        """

        if isinstance(xml_file, pathlib.Path):
            xml_file = xml_file.absolute

        if self._nmap_bin:
            nmap_bin = self._nmap_bin
        else:
            nmap_bin = 'nmap'

        _, error_buff = self._execute_nmap(self._split_command('{} --resume {}'.format(nmap_bin, xml_file)))

        # Resume should not be checked through output_buff, since a --resume command may not any output at all.
        if not error_buff:
            return self._xml_parser.parse_file(xml_file)
        else:
            raise NmapScanError(error_buff.decode('utf8'))

    def from_file(self, xml_file: Union[pathlib.Path,str]) -> NmapScanResult:
        """ Imports an existing XML file and returns a scan result
        
        :param xml_file: String representing the file's path
        """

        return self._xml_parser.parse_file(xml_file)


#!/usr/bin/env python3

# Copyright (c) 2019 Christian Barral

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# This module contains the XML parser used to transform 
# pure Nmap XML output into Python objects.

import pathlib

import xml.etree.ElementTree as ET

from typing import Union

from .elements import Host, Port, Service, OperatingSystem, Hop
from .results import NmapScanResult
from .exceptions import XMLParsingError


class XMLParser:
    """ Used to parse Nmap outputs into Python objects.

    Parsing can occur from a plain string to a phisical XML file on
    the file system. All the information will be encapsulated into different
    Python objects. 

    Attributes:
        _xml_tree: Tree containing all the XML information
    """

    __slots__ = ('_xml_tree',)

    def __init__(self):
        self._xml_tree = None
    
    @property
    def xml_tree(self):
        return self._xml_tree

    def parse_file(self, file_path: Union[pathlib.Path,str]):
        """ Parse a XML file in the system.

        :param file_path: Path from the file as a String or a Path object.
        :raises FileNotFoundError: If file is not found and as_boolean is False.
        :returns: Parsed file as NmapScanResult.
        """

        if isinstance(file_path, pathlib.Path):
            file_path = file_path.absolute
        with open(file_path) as f:
            return self._parse(f.read())

    def parse_plain(self, plain_text: str):
        """ Parse a plain string that contains the XML.

        :param plain_text: Plain string containing the data.
        :returns: Parsed string as NmapScanResult
        """

        return self._parse(plain_text)

    def _parse(self, text: str) -> NmapScanResult:
        """ Contains the logic for parsing Nmap XML output from a string.

        :param text: Text to parse.
        :returns: Scan result
        """

        try:
            self._xml_tree = ET.fromstring(text)
        except ET.ParseError as e:
            raise XMLParsingError('Cannot parse Nmap XML output: {}'.format(e)) from None
        
        # Parse general scan information
        general_info = {}
        # Loop through every attribute from the root element
        for attribute, value in self._xml_tree.attrib.items():
            if attribute == 'scanner':
                general_info['scanner'] = value
            elif attribute == 'args':
                general_info['arguments'] = value
            elif attribute == 'start':
                general_info['start_timestamp'] = value
                general_info['start_datetime'] = value
            elif attribute == 'version':
                general_info['version'] = value
        
        # Loop through all elements on the <finished> element
        for attribute, value in self._xml_tree.find('.//finished').attrib.items():
            if attribute == 'time':
                general_info['end_timestamp'] = value
                general_info['end_datetime'] = value
            elif attribute == 'elapsed':
                general_info['elapsed'] = value
            elif attribute == 'summary':
                general_info['summary'] = value
            elif attribute == 'exit':
                general_info['exit_status'] = value
        
        # Loop through all elements on <hosts> element
        for attribute, value in self._xml_tree.find('.//hosts').attrib.items():
            if attribute == 'up':
                general_info['hosts_up'] = value
            elif attribute == 'down':
                general_info['hosts_down'] = value
            elif attribute == 'total':
                general_info['num_hosts'] = value

        scan_info = {}

        # Parse all the <scaninfo> elements
        for element in self._xml_tree.findall('.//scaninfo'):
            
            current_scan_tag_attributes = element.attrib
            scan_info[current_scan_tag_attributes['protocol']] = {
                'type': current_scan_tag_attributes['type'],
                'numservices': current_scan_tag_attributes['numservices'],
                'services': current_scan_tag_attributes['services']
            }

        general_info['scan_info'] = scan_info

        # Parse the <verbose> and <debugging> tags if they exist
        verb_elem = self._xml_tree.find('.//verbose')
        if verb_elem is not None:
            general_info['verbose'] = verb_elem.attrib['level']

        debug_elem = self._xml_tree.find('.//debugging')
        if debug_elem is not None:
            general_info['debug'] = debug_elem.attrib['level']

        scan_result = NmapScanResult(**general_info)

        # Loop through every <host> element, which contains very host scan result.
        for host in self._xml_tree.findall('.//host'):
            host_info = {
                'start_time': host.attrib['starttime'],
                'end_time': host.attrib['endtime']
            }
            status_element = host.find('status')
            if status_element is None:
                raise XMLParsingError('Could not get status from host')
            host_info['state'] = status_element.attrib['state'] 
            host_info['reason'] = status_element.attrib['reason'] 
            host_info['reason_ttl'] = status_element.attrib['reason_ttl'] 
            address_items = host.findall('.//address')
            if not address_items:
                raise XMLParsingError('Could not be able to parse host address')
            
            # Parse IPv4 and IPv6 if exist
            for addr in address_items:
                if addr.attrib['addrtype'] == 'ipv4':
                    host_info['ipv4'] = addr.attrib['addr']
                elif addr.attrib['addrtype'] == 'ipv6':
                    host_info['ipv6'] = addr.attrib['addr']
            
            if 'ipv4' not in host_info and 'ipv6' not in host_info:
                raise XMLParsingError('Cannot parse host that has no IPv4 nor IPv6 address')

            # Parse hostnames
            hostnames_element = host.find('hostnames')
            if hostnames_element is not None:
                host_info['hostnames'] = {}
                for hostname_element in hostnames_element:
                    host_info['hostnames'][hostname_element.attrib['name']] = hostname_element.attrib['type']

            # Get OS fingerprint
            os_fingerprint_element = host.find('.//osfingerprint')
            if os_fingerprint_element is not None:
                host_info['fingerprint'] = os_fingerprint_element.attrib['fingerprint']

            # Instatiate the host
            host_instance = Host(**host_info)

            # Parse all ports
            scan_info = host.find('ports')
            if scan_info is not None:
                for port in scan_info.findall('port'):
                    port_info = {
                        'protocol': port.attrib['protocol'],
                        'number': port.attrib['portid']
                    }
                    
                    state_element = port.find('state')
                    if state_element is None:
                        raise XMLParsingError('Cannot find state element from port')
                    port_info['state'] = state_element.attrib['state']
                    port_info['reason'] = state_element.attrib['reason']
                    port_info['reason_ttl'] = state_element.attrib['reason_ttl']

                    # Create the port object
                    port_instance = Port(**port_info)

                    # Parse service information
                    service_info = {'port': port_info['number']}
                    service_element = port.find('service')
                    if service_element is not None:
                        service_info['name'] = service_element.attrib['name']
                        try:
                            service_info['product'] = service_element.attrib['product']
                        except KeyError:
                            service_info['product'] = None
                        try:
                            service_info['version'] = service_element.attrib['version']
                        except KeyError:
                            service_info['version'] = None
                        try:
                            service_info['extrainfo'] = service_element.attrib['extrainfo']
                        except KeyError:
                            service_info['extrainfo'] = None
                        try:
                            service_info['tunnel'] = service_element.attrib['tunnel']
                        except KeyError:
                            service_info['tunnel'] = None
                        try:
                            service_info['method'] = service_element.attrib['method']
                        except KeyError:
                            service_info['method'] = None
                        try:
                            service_info['conf'] = service_element.attrib['conf']
                        except KeyError:
                            service_info['conf'] = None
                        
                        service_info['cpes'] = []

                        # Get CPEs
                        for cpe_item in service_element.findall('cpe'):
                            service_info['cpes'].append(cpe_item.text)

                        # Bind the service instance with the port instance
                        service_instance = Service(**service_info)

                        for script in port.findall('script'):
                            service_instance._add_script(script.attrib['id'], script.attrib['output'])

                        port_instance._add_service(service_instance)

                    # Bind the port instance to the current host
                    host_instance._add_port(port_instance)

            os_root_element = host.find('os')

            # Add OS information
            if os_root_element is not None:

                for os_element in os_root_element.findall('osmatch'):
                    os_info = {}
                    os_info['name'] = os_element.attrib['name']
                    os_info['accuracy'] = os_element.attrib['accuracy']
                    matches = []
                    for os_match_element in os_element.findall('osclass'):
                        match_info = {}
                        for attrib_name in ('type', 'vendor', 'osfamily', 'osgen'):
                            if attrib_name == 'osfamily':
                                target_param = 'family'
                            elif attrib_name == 'osgen':
                                target_param = 'generation'
                            else:
                                target_param = attrib_name
                            try:
                                match_info[target_param] = os_match_element.attrib[attrib_name]
                            except KeyError:
                                match_info[target_param] = None
                        
                        match_info['cpe'] = None

                        cpe_element = os_match_element.find('cpe')
                        if cpe_element is not None:
                            match_info['cpe'] = cpe_element.text
                        
                        matches.append(match_info)
                    
                    os_info['matches'] = matches
                    os_instance = OperatingSystem(**os_info)
                    host_instance._add_os(os_instance) 
            
            # Parse traceroute
            trace_element = host.find('trace')
            if trace_element is not None:
                hops = []
                for hop in trace_element.findall('hop'):
                    hop_info = {}
                    try:
                        hop_info['host'] = hop.attrib['host']
                    except KeyError:
                        hop_info['host'] = None
                    try:
                        hop_info['ttl'] = hop.attrib['ttl']
                    except KeyError:
                        hop_info['ttl'] = None
                    try:
                        hop_info['rtt'] = hop.attrib['rtt']
                    except KeyError:
                        hop_info['rtt'] = None
                    try:
                        hop_info['ip'] = hop.attrib['ipaddr']
                    except KeyError:
                        hop_info['ip'] = None

                    hops.append(Hop(**hop_info))
                
                host_instance._add_hops(*hops)

            # Parse host scripts
            hostscript_element = host.find('hostscript')
            if hostscript_element:
                for script_element in hostscript_element.findall('script'):
                    host_instance._add_script(script_element.attrib['id'], script_element.attrib['output'])

            scan_result._add_hosts(host_instance)

        return scan_result
        


from setuptools import setup

with open('README.md') as f:
    setup(
        name='nmapthon2',
        version='0.1.000',
        packages=['nmapthon2'],
        url='https://github.com/cblopez/nmapthon',
        license='GLPv3',
        author='cblopez',
        author_email='cbarrallopez@gmail.com',
        description='The evolution of nmapthon, an Nmap automation library for humans.',
        long_description=f.read(),
        long_description_content_type='text/markdown',
        classifiers=[
            'Development Status :: 4 - Beta',
            'Programming Language :: Python :: 3 :: Only',
            'Topic :: System :: Networking',
            'Topic :: Software Development :: Libraries :: Python Modules'
        ],
        keywords=['python', 'python3', 'nmap', 'module', 'scan', 'nse', 'port', 'service', 'async', 'network']
    )


#!/usr/bin/env python

# Copyright (c) 2019 Christian Barral

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import socket
import struct

from collections.abc import Iterable

from .exceptions import InvalidPortError, MalformedIpAddressError

_BASE_IP_REGEX = '(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])'
_SINGLE_IP_ADDRESS_REGEX = re.compile('^{}$'.format(_BASE_IP_REGEX))
_IP_ADDRESS_WITH_CIDR_REGEX = re.compile('^{}/([0-9]|[1-2][0-9]|3[0-2])$'.format(_BASE_IP_REGEX))
_IP_RANGE_REGEX = re.compile('^{}-{}$'.format(_BASE_IP_REGEX, _BASE_IP_REGEX))
_OCTECT_RANGE_REGEX = '(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))'
_PARTIAL_IP_RANGE_REGEX = re.compile('{}(-{})?\.{}(-{})?\.{}(-{})?\.{}(-{})?'.format(*[_OCTECT_RANGE_REGEX for _ in range(8)]))


def valid_port(port) -> bool:
    """Checks if a given value might be a port. 
    Must be between 1 and 65535, both included.

        :param port: Port candidate
        :returns: True if is valid, False if not
    """
    try:
        int_port = int(port)
    except ValueError:
        return False

    return 0 < int_port < 65536


def ports_to_list(ports: str):
    """ Returns a list containing all ports specified in 
    a nmap-format port string.

    :param ports: String that specifies the ports to scan
    :return: List containing the ports to scan as Strings.
    :raises: InvalidPortError

    Example:
        ports               return
        '10,20,30'          ['10', '20', '30']
        '10-13'             ['10'. '11', '12', '13']
        '80, 81-83'         ['80', '81', '82', '83']
    """
    # Create empty list and delete any blank spaces
    port_list = []
    # Delete blank spaces
    ports_string = ports.replace(' ', '')

    # For every comma separated block
    for split_ports in ports_string.split(','):
        # If there is a range indicator.
        if '-' in split_ports:
            # Split the range
            port_range = split_ports.split('-')
            # Cast to integer the starting port range number.
            try:
                first_port_range = int(port_range[0])
            # If ValueError, non valid port, format the InvalidPortError message.
            except ValueError:
                first_port = port_range[0] if len(port_range[0]) else 'None'
                raise InvalidPortError('Invalid starting port range: {}'.format(first_port)) from None
            # Cast ending port range
            try:
                last_port_range = int(port_range[1]) + 1
            # If IndexError, no ending port range was specified.
            except IndexError:
                raise InvalidPortError('End of port range in {}- not specified'.format(port_range[0])) from None
            # If ValueError, invalid ending for port range.
            except ValueError:
                raise InvalidPortError('Invalid ending port range: {} '.format(port_range[1])) from None
            # For every port in the range calculated
            for single_port in range(first_port_range, last_port_range):
                # If valid port, add to list
                if valid_port(single_port):
                    port_list.append(single_port)
                # If invalid, raise Exception
                else:
                    raise InvalidPortError('{} is not a valid port'.format(single_port))

        # If no range indicators, guess individual port
        else:
            # If split port has length
            if len(split_ports):
                # Cast to integer value
                try:
                    integer_parsed_port = int(split_ports)
                # If ValueError, malformed
                except ValueError:
                    raise InvalidPortError('Invalid port: {}'.format(split_ports)) from None
                # If is a valid port, append it to list
                if valid_port(integer_parsed_port):
                    port_list.append(integer_parsed_port)
                # If invalid, raise Error
                else:
                    raise InvalidPortError('{} is not a valid port.'.format(integer_parsed_port))
    
    return sorted(list(set(port_list)))

def extend_port_list(port_list: Iterable) -> list:
    """ Transforms a port list with single ports and/or port ranges 
    into a single list with no duplicates.

    :param port_list: Port list to parse
    :returns: List of single and unique ports
    """

    # To STR
    port_list = list(map(str, port_list))

    new_port_list = []
    for i in list(map(ports_to_list, port_list)):
        new_port_list.extend(i)
    
    return list(set(new_port_list))


def ports_to_str(port_list: Iterable) -> str:
    """ Parse a list of int/str ports into a single str containing a 
    port range that nmap can understand.

        :param port_list: List of ports
        :return: String representing the ports in nmap syntax
    """

    # Get unique list of single ports
    new_port_list = extend_port_list(port_list)

    # If not all ports are valid ports, raise NmapScanError
    if not all(valid_port(p) for p in new_port_list):
        raise InvalidPortError('Ports must be between 0 and 65536') from None

    # Sort ports in ascending order
    sorted_ports = sorted(new_port_list)
    # Instantiate port string
    port_string = ''
    # Instantiate last port variable
    last_port = -2
    # Loop through sorted_ports list, bust access the ports by position.
    # Used to know when the list is about to finish
    i = 0
    while i < len(sorted_ports):
        # If the current port is not the immediate next one
        if sorted_ports[i] != (last_port + 1):
            port_string += str(sorted_ports[i])
            # Set last_port to current port
            last_port = sorted_ports[i]
            # Append a comma if the port is not the last one
            if i != (len(sorted_ports) - 1) and sorted_ports[i + 1] != (sorted_ports[i] + 1):
                port_string += ','
            # Add one to i
            i += 1
        # If the current port is the previous port plus 1
        else:
            # Add a dash (range indicator)
            port_string += '-'
            # Set last_port to current port
            last_port = sorted_ports[i]
            # Add one to i
            i += 1
            # keep going if next ports are immediate next
            while i < len(sorted_ports) and sorted_ports[i] == (last_port + 1):
                last_port = sorted_ports[i]
                i += 1
            # On exit of while loop, range ends, write last port range
            port_string += str(last_port)
            # Append a comma if the port is not the last one
            if i != len(sorted_ports):
                port_string += ','

    return port_string

def valid_ip(ip_address: str) -> bool:
    """Checks if a given IP address is correctly formed.

        :param ip_address: IP address to check
        :return: True if it is a valid IP Address, False if not
    """

    # Return True if matches, False if not.
    if _SINGLE_IP_ADDRESS_REGEX.fullmatch(ip_address):
        return True
    else:
        return False


def ip_range(starting_ip: str, ending_ip: str) -> list:
    """ Calculates a list of IPs between two given.

        :param starting_ip: Range starting IP address
        :param ending_ip: Range ending IP address
        :returns: List of ports between two given IPs
    """

    # Create a list contaning the 4 octets from both IP address in decimal format.
    split_starting_ip = list(map(int, starting_ip.split('.')))
    split_ending_ip = list(map(int, ending_ip.split('.')))
    # Create list of IPs to return, starting with the first one.
    ip_range = [starting_ip]

    # Execute algorithm. While you can add one to the most on the right octet, keep going
    # and add. If the 4 octets are named from 3 to 0 from left to right: when octet N is 255,
    # set octet N to 0 and add one to octet N+1
    while split_starting_ip != split_ending_ip:
        split_starting_ip[3] += 1
        for i in [3, 2, 1]:
            if split_starting_ip[i] == 256:
                split_starting_ip[i] = 0
                split_starting_ip[i - 1] += 1
        # Reformat to IP address-like string.
        current_ip = '.'.join(map(str, split_starting_ip))
        ip_range.append(current_ip)

    return ip_range


def partial_ip_range(ip_addr: str) -> list:
    """ Calculates the list of IP address from a partial ranged IP expression.

        :param ip_addr: IP Address from where to extract the IPs
        :returns: List of IPs in partial range
    """

    # Split by dots
    split_ip = ip_addr.split('.')
    # IPs to return
    ips=[]
    # List to store each part range
    partial_ranges = []
    # For each partial IPs part
    for i in split_ip:
        # If its a range
        if '-' in i:
            # Extract the list of numbers between
            partial_range = i.split('-')
            try:
                start = int(partial_range[0])
            except ValueError:
                raise MalformedIpAddressError('Invalid start of range, expected number but got : {}'.format(partial_range[0]))
            try:
                end = int(partial_range[1])
            except ValueError:
                raise MalformedIpAddressError('Invalid start of range, expected number but got : {}'.format(partial_range[1]))

            if not 0 <= start <= end <= 255:
                raise MalformedIpAddressError('Start range must be lower than end range, and both between 0 adn 255')

            partial_ranges.append(list(range(start, end + 1)))

        # If not, add a list with a single element
        else:
            partial_ranges.append([i])
    
    # Combine them all
    # TODO: Beautify this
    for one in partial_ranges[0]:
        for two in partial_ranges[1]:
            for three in partial_ranges[2]:
                for four in partial_ranges[3]:
                    ips.append('{}.{}.{}.{}'.format(one, two, three, four))
    
    return ips


def dispatch_network(network: str) -> list:
    """ Creates a list of all the IP address inside a network with 
    it's net-mask in CIDR format.

        :param network: Network IP address and /net-mask to dispatch
        :returns: List of every IP on a network range
        :raises: MalformedIPAddressError
    """

    # List to return
    ip_addresses = []

    # Delete blank spaces and split IP Address and netmask in CIDR format.
    ip_address_netmask = network.replace(' ', '').split('/')
    # If not split in two parts, raise Exception.
    if len(ip_address_netmask) != 2:
        raise MalformedIpAddressError('Invalid network to dispatch: {}.'
                                      ' Need an IP address and CIDR Mask like 192.168.1.0/24'
                                      .format(ip_address_netmask))

    # IP Address is the first part
    ip_address = ip_address_netmask[0]

    # CIDR is the second part
    try:
        cidr = int(ip_address_netmask[1])
    # If cannot convert to integer, raise Exception
    except ValueError:
        raise MalformedIpAddressError('Invalid CIDR format: {}'.format(ip_address_netmask[1])) from None

    # If netmask not between 0 and 32, included, raise Exception
    if not 0 <= cidr <= 32:
        raise MalformedIpAddressError('Out of range CIDR: {}'.format(cidr))

    # If invalid IP address, raise Exception
    if not valid_ip(ip_address):
        raise MalformedIpAddressError('Invalid network IP: {}.'.format(ip_address))

    # Combination from struct and socket for binary formatting and bit level operations.
    # Getting every IP address inside a network range (established by netmask).
    host_bits = 32 - cidr
    aux = struct.unpack('>I', socket.inet_aton(ip_address))[0]
    start = (aux >> host_bits) << host_bits
    end = start | ((1 << host_bits) - 1)

    for ip in range(start, end):
        ip_addresses.append(socket.inet_ntoa(struct.pack('>I', ip)))

    # Return every IP address but not Network Address
    return ip_addresses[1:]


def targets_to_list(targets: str) -> list:
    """ Returns a list containing all targets specified for the scan.

        :param targets: String that specifies the targets to scan
        :return: List containing the targets to scan as Strings.
        :raises: MalformedIPAddressError

        Example:
            targets                             return
            '192.168.1.1 192.168.1.2'          ['192.168.1.1', '192.168.1.2']
            '192.168.1.1-192.168.1.3'           ['192.168.1.1', '192.168.1.2', '192.168.1.3']
            '192.168.1.0/30'                    ['192.168.1.1', '192.168.1.2']

        note:
            If network/cidr mask is specified, both Network address and broadcast address will be omitted.
    """

    # List to return
    target_list = []

    # For each block split by a comma.
    for split_target in targets.split(' '):
        if not split_target:
            continue
        # If range indicator
        if _IP_RANGE_REGEX.fullmatch(split_target):
            # Split range
            ip_range_list = split_target.split('-')
            # Get starting IP address from range
            starting_ip = ip_range_list[0]
            # If not a valid IP address, raise Error
            if not valid_ip(starting_ip):
                raise MalformedIpAddressError('Invalid starting IP range: {}'.format(starting_ip))
            # Get Ending IP address from range
            ending_ip = ip_range_list[1]
            # If not valid IP address, raise Error
            if not valid_ip(ending_ip):
                raise MalformedIpAddressError('Invalid ending IP range: {}'.format(ending_ip))
            # For every IP in range, add to list if valid IP. If not, raise Exception.
            for single_target_in_range in ip_range(starting_ip, ending_ip):
                if valid_ip(single_target_in_range):
                    target_list.append(single_target_in_range)
                else:
                    raise MalformedIpAddressError('Invalid IP Address: {}'.format(single_target_in_range))
        # If a slash is found, guess a network mask
        elif _IP_ADDRESS_WITH_CIDR_REGEX.fullmatch(split_target):
            # Extend the list for dispatching the network
            target_list.extend(dispatch_network(split_target))

        # If partial IP addresses
        elif _PARTIAL_IP_RANGE_REGEX.fullmatch(split_target):
            target_list.extend(partial_ip_range(split_target))

        # If it reaches here, guess single IP. Add to list or raise Error if malformed.
        else:
            target_list.append(split_target)

    # Return the sorted list. List is sorted by IP address. Ej: 192.168.1.12 > 192.168.1.9
    return list(set(target_list))

